import numpy as np

from geant4_pybind import G4ThreeVector

from g4camp.IO.gQuant import gQuant

class gParticles():
    
    def __init__(self, label: str = ''):
        
        self._label   = label
        self._counter = 0
    
        self.uid           = gQuant('uid',int)
        self.parent_uid    = gQuant('parent_uid',int)
        self.pdgid         = gQuant('pdgid',int)
        self.x_m           = gQuant('x_m',float)
        self.y_m           = gQuant('y_m',float)
        self.z_m           = gQuant('z_m',float)
        self.t_ns          = gQuant('t_ns',float)
        self.dir_x         = gQuant('dir_x',float)
        self.dir_y         = gQuant('dir_y',float)
        self.dir_z         = gQuant('dir_z',float)
        self.Etot_GeV      = gQuant('Etot_GeV',float)
        self.process_name  = gQuant('process_name','|S10')
        
        self.quantities = (self.uid,self.parent_uid,self.pdgid,
                    self.x_m,self.y_m,self.z_m,self.t_ns,
                    self.dir_x,self.dir_y,self.dir_z,self.Etot_GeV,
                    self.process_name)
    
    @property
    def label(self):
        return self._label
    
    @property
    def counter(self):
        return self._counter
    
    @counter.setter
    def counter(self, new_count):
        self._counter = new_count
    
    @property
    def pos_m(self):
        return np.column_stack((self.x_m,
                                self.y_m,
                                self.z_m))
    
    @property
    def dir(self):
        return np.column_stack((self.dir_x,
                                self.dir_y,
                                self.dir_z))
    
    @property
    def unique_data(self):
        return len(set(self.uid.value))
    
    def set_data(self, new_uid: int, new_parent_uid: int, new_pdgid: int, \
                 new_pos_m: G4ThreeVector, new_t_ns: float,
                 new_dir: G4ThreeVector, new_Etot_GeV: float, \
                 new_process_name: bytes = None) -> None:
        
        self.uid.value           = new_uid
        self.parent_uid.value    = new_parent_uid
        self.pdgid.value         = new_pdgid
        
        self.x_m.value           = new_pos_m.x
        self.y_m.value           = new_pos_m.y
        self.z_m.value           = new_pos_m.z
        self.t_ns.value          = new_t_ns
        
        self.dir_x.value         = new_dir.x
        self.dir_y.value         = new_dir.y
        self.dir_z.value         = new_dir.z
        self.Etot_GeV.value      = new_Etot_GeV
        
        if new_process_name:
            self.process_name.value = new_process_name
        
        self.counter += 1
    
    def get_named_data(self) -> np.ndarray:
        
        quantities = tuple(q for q in self.quantities if q.has_value())
        if not quantities:
            return np.empty(shape=(0),dtype=None)
        data_type  = list((q.label,q.dtype) for q in quantities)
        data_list  = np.column_stack([q.value for q in quantities])
        named_data = np.array([tuple(_) for _ in data_list], dtype=data_type)
        named_data = named_data[np.lexsort((named_data['t_ns'], named_data['uid']))]
        
        return named_data


from geant4_pybind import *

class UserPhysicsList(G4VUserPhysicsList):
    
    def __init__(self) -> None:
        super().__init__()
    
    def ConstructParticle(self) -> None:
        self.electron = G4Electron.Definition()
        G4Positron.Definition()
        G4Gamma.Definition()
        G4MuonMinus.Definition()
        G4MuonPlus.Definition()
        G4PionMinus.Definition()
        G4PionPlus.Definition()
    
    def ConstructProcess(self) -> None:
        proc = G4ProcessTable
        G4ComptonScattering.GetProcessName()
        proc.SetProcessActivation(G4ProcessType.fElectromagnetic, self.electron, True)


import numpy as np

class gQuant():
    
    def __init__(self, label: str, dtype: np.dtype):
        
        self._label = label
        self._dtype = dtype
        
        self.__capacity = 100_000
        self.__size     = 0
        
        self._value = np.empty(shape=self.__capacity, dtype=self._dtype)
    
    @property
    def label(self) -> str:
        return self._label
    
    @property
    def dtype(self) -> type:
        return self._dtype
    
    @property
    def value(self) -> dtype:
        if self._value is None:
            raise ValueError('Value has not been initialized!')
        if np.size(self._value) != self.__size:
            self._value = self._value[:self.__size]
        return self._value
    
    @value.setter
    def value(self, new_value: dtype) -> None:
        if self.__size == self.__capacity:
            self.__capacity *= 2
            data_reshape = np.empty(shape=self.__capacity, dtype=self._dtype)
            data_reshape[:self.__size] = self._value
            self._value = data_reshape
        
        self._value[self.__size] = new_value
        
        self.__size += 1
    
    def has_value(self) -> bool:
        return self.__size > 0


import numpy as np

from geant4_pybind import *

class SteppingAction(G4UserSteppingAction):

  def __init__(self, app):
    
    super().__init__()
    
    self.app = app
    self.data_buffer = app.data_buffer
    
    n = self.app.refractive_index
    self.cerenkov_threshold_e = G4ParticleDefinition.GetPDGMass(G4Electron.Electron())/GeV/np.sqrt(1.-1./n**2)
  
  def UserSteppingAction(self, step: G4Step):
    
    pre_step_point = step.GetPreStepPoint()
    Etot1 = pre_step_point.GetTotalEnergy()/GeV
    
    # Save every track point
    if Etot1 >= self.cerenkov_threshold_e and Etot1 >= self.app.E_lower_edge:
      # keep step end point
      track    = step.GetTrack()   
      particle = track.GetDefinition()
      
      uid           = track.GetTrackID()
      parent_uid    = track.GetParentID()
      pdgid         = particle.GetPDGEncoding()
      pos_m         = track.GetPosition()/m
      t_ns          = track.GetGlobalTime()/ns
      Etot_GeV      = track.GetTotalEnergy()/GeV
      step_length_m = track.GetStepLength()/m
      if self.app.save_process_name:
        post_step_point = step.GetPostStepPoint()
        process_name = post_step_point.GetProcessDefinedStep().GetProcessName()
        self.data_buffer.tracks = [uid, parent_uid, pdgid, pos_m, t_ns,
                                   Etot_GeV, step_length_m, process_name]
      else:
        self.data_buffer.tracks = [uid, parent_uid, pdgid, pos_m,
                                   t_ns, Etot_GeV, step_length_m]


import numpy as np

from geant4_pybind import G4ThreeVector

from g4camp.IO.gQuant import gQuant

class gPhotons():
    
    def __init__(self, label: str = ''):
        
        self._label   = label
        self._counter = 0
    
        self.parent_uid = gQuant('parent_uid',int)
        self.x_m        = gQuant('x_m',float)
        self.y_m        = gQuant('y_m',float)
        self.z_m        = gQuant('z_m',float)
        self.t_ns       = gQuant('t_ns',float)
        self.dir_x      = gQuant('dir_x',float)
        self.dir_y      = gQuant('dir_y',float)
        self.dir_z      = gQuant('dir_z',float)
        self.wl_nm      = gQuant('wl_nm',float)
        
        self.quantities = (self.parent_uid,
                           self.x_m,self.y_m,self.z_m,self.t_ns,
                           self.dir_x,self.dir_y,self.dir_z,
                           self.wl_nm)
    
    @property
    def label(self):
        return self._label
    
    @property
    def counter(self):
        return self._counter
    
    @counter.setter
    def counter(self, new_count):
        self._counter = new_count
    
    @property
    def pos_m(self):
        return np.column_stack([self.x_m,
                                self.y_m,
                                self.z_m])
    
    @property
    def dir(self):
        return np.column_stack([self.dir_x,
                                self.dir_y,
                                self.dir_z])
    
    @property
    def unique_data(self):
        return self.counter
    
    def set_data(self, new_parent_uid: int, new_pos_m: G4ThreeVector, new_t_ns: float, \
                 new_dir: G4ThreeVector, new_wavelength_nm: float) -> None:
        
        self.parent_uid = new_parent_uid
        
        self.x_m        = new_pos_m.x
        self.y_m        = new_pos_m.y
        self.z_m        = new_pos_m.z
        self.t_ns       = new_t_ns
        
        self.dir_x      = new_dir.x
        self.dir_y      = new_dir.y
        self.dir_z      = new_dir.z
        
        self.wl_nm      = new_wavelength_nm
        
        self.counter += 1
    
    def get_named_data(self) -> np.ndarray:
        
        quantities = tuple(q for q in self.quantities if q.has_value())
        if not quantities:
            return np.empty(shape=(0),dtype=None)
        data_type  = list((q.label,q.dtype) for q in quantities)
        data_list  = np.column_stack([q.value for q in quantities])
        named_data = np.array([tuple(_) for _ in data_list], dtype=data_type)
        
        return named_data


from g4camp.g4camp import g4camp
import time
import numpy as np
import h5py
import sys
import configargparse
import logging

log = logging.getLogger('run_g4camp')
logformat='[%(name)12s ] %(levelname)8s: %(message)s'
logging.basicConfig(format=logformat)

def report_timing(func):
    def func_wrapper(*args, **kwargs):
        tic = time.time()
        result = func(*args, **kwargs)
        toc = time.time()
        dt = toc - tic
        log .info(f'{func.__name__}'.ljust(40)+f'{dt:6.3f} seconds'.rjust(30))
        return result
    return func_wrapper


def parse_arguments():
    p = configargparse.get_argument_parser()
    p.add_argument('-n', type=int, default=2, help="number of events")
    p.add_argument('-o', type=str, default='output.h5', help="output file name")
    p.add_argument('--physics', type=str, default='Custom', help="choose physics list", choices=('Custom', 'FTFP_BERT', 'QGSP_BERT', 'QGSP_BIC'))
    p.add_argument('--mode_custom_physlist', type=str, choices=('all_phys', 'em_cascade', 'fast'), default='all_phys', help="choose available physics set")
    p.add_argument('--enable_optics', action='store_true', help="enable optics")
    p.add_argument('-p', '--primary-generator', type=str, choices=('gun','gps'), default='gun', help="type of primary generator type")
    p.add_argument('--gun_particle', type=str, default='mu-', help="gun particle type, e.g. e-, e+, gamma, mu-, pi0 etc.")
    p.add_argument('--gun_energy_GeV', type=float, default=100., help="gun particle energy in GeV")
    p.add_argument('--gun_position_m', type=float, nargs=3, default=[0.,0.,0.], help="gun position in meters")
    p.add_argument('--gun_direction', type=float, nargs=3, default=[0.,0.,1.], help="gun direction X, Y, Z components")
    p.add_argument('--gps_macro', type=str, default='muons.mac', help="macro files with Geant4 GPS commands to congigure particle source")
    p.add_argument('--event_max_energy_GeV', type=float, default=100., help="maximum energy of particles in an event in GeV")
    p.add_argument('--e_min_energy_GeV', type=float, default=1e-7, help="minimum energy of electrons/positrons in GeV")
    p.add_argument('--muhad_min_energy_GeV', type=float, default=1e-7, help="minimum energy of muons/hadrons in GeV")
    p.add_argument('--step_function_m', type=float, nargs=2, default = (0.2, 1e-4), help='Step function parameters (the second parameter is specified in meters)')
    p.add_argument('--skip_mode', type=str, choices=('cut','fraction'), default='cut', help='choice of the option for cutting secondary particles: (i) cut - through absolute values, (ii) fraction - through the fraction of the energy of the initial particle')
    p.add_argument('--skip_min', type=float, default=0.002, help="minimal of particle energy to skip")
    p.add_argument('--skip_max', type=float, default=0.01, help="maximal of particle energy to skip")
    p.add_argument('--ene_lower_edge', type=float, default=0., help="minimum total energy of the particles in the simulation")
    p.add_argument('--random_seed', type=int, default=1, help="random seed")
    p.add_argument('--photon_suppression', type=int, default=10, help="photon suppression factor")
    p.add_argument('--refractive_index', type=float, default=1.34, help="refractive index of a transparent medium")
    p.add_argument('--det_height', type=float, default=1500., help="(cylindrical) detector volume height (in meters)")
    p.add_argument('--det_radius', type=float, default=1500., help="(cylindrical) detector volume radius (in meters)")
    p.add_argument('--save_process_name', action='store_true', help='')
    p.add_argument('-l', '--log-level', type=str, choices=('deepdebug', 'debug', 'info', 'warning', 'error', 'critical'), default='info', help='logging level')

    opts = p.parse_args()
    #
    log.setLevel(opts.log_level.upper())
    log.info("----------")
    log.info(p.format_help())
    log.info("----------")
    log.info(p.format_values())    # useful for logging where different settings came from
    return opts


@report_timing
def run(app, n_events=10, ofname="output.h5"):
    h5file = h5py.File(ofname, "w")  # create empty file
    h5file.close()
    for ievt, data in enumerate(app.run(n_events)):
        h5file = h5py.File(ofname, "a")
        g = h5file.create_group(f"event_{ievt}")
        g.create_dataset('particles', data=data.particles.get_named_data())
        g.create_dataset('photons', data=data.photons.get_named_data())
        g.create_dataset('tracks', data=data.tracks.get_named_data())
        h5file.close()
        #log.info(f" Event #{ievt}/{n_events}")
        log.info(f"   Number of particles:    {data.particles.unique_data}")
        log.info(f"   Number of tracks:       {data.tracks.unique_data}")
#        log.info(f"   Number of track points: {len(data.tracks.quantities['uid'].value)}")
        log.info(f"   Number of photons:      {data.photons.unique_data}")
        db = app.data_buffer
        #log.info(f"   Number of particles / tracks points / photons : {len(db.particles):6.0f} / {len(db.tracks):6.0f} / {len(db.photons):6.0f}")
        log.info(" ")


def main():
    opts = parse_arguments()
    physics = opts.physics
    mode_physlist = opts.mode_custom_physlist
    optics = opts.enable_optics
    #
    if opts.primary_generator == 'gun':
        gun_args = {'particle': opts.gun_particle, 
                    'energy_GeV': opts.gun_energy_GeV, 
                    'position_m': opts.gun_position_m,
                    'direction': opts.gun_direction}
        app = g4camp(physics=physics, mode_physlist=mode_physlist, optics=optics, primary_generator='gun', gun_args=gun_args)
    elif opts.primary_generator == 'gps':
        app = g4camp(physics=physics, mode_physlist=mode_physlist, optics=optics, primary_generator='gps')
        app.setGPSMacro(opts.gps_macro)
    else:
        log.error(f"Wrong primary type '{opts.primary_generator}'")
    app.log = log
    app.setRandomSeed(int(opts.random_seed))
    app.setSkipMinMax(opts.skip_mode, opts.skip_min, opts.skip_max)
    app.setPhotonSuppressionFactor(opts.photon_suppression)
    app.setDetectorHeight(opts.det_height)
    app.setDetectorRadius(opts.det_radius)
    app.SaveProcessName(opts.save_process_name)
    app.configure()
    run(app, n_events=int(opts.n), ofname=opts.o)
    log.info("All done!")


if __name__ == "__main__":
    main()




from geant4_pybind import *
import numpy as np

class StackingAction(G4UserStackingAction):

    def __init__(self, app):
        super().__init__()
        
        self.app = app
        self.data_buffer = app.data_buffer
        
        n = self.app.refractive_index
        self.cerenkov_threshold_e = G4ParticleDefinition.GetPDGMass(G4Electron.Electron())/GeV/np.sqrt(1.-1./n**2)

    def ClassifyNewTrack(self, track: G4Track):
        
        if track.GetDefinition().GetParticleType() == 'nucleus':
            return G4ClassificationOfNewTrack.fKill
        
        particle = track.GetDefinition()
        uid      = track.GetTrackID()
        Etot_GeV = track.GetTotalEnergy()/GeV
        
        if track.GetDefinition() == G4OpticalPhoton.OpticalPhoton():
            if self.app.ph_counter % self.app.ph_suppression_factor == 0:
                parent_uid = track.GetParentID()
                pos_m      = track.GetPosition()/m
                t_ns       = track.GetGlobalTime()/ns
                dir        = track.GetMomentumDirection()
                wl_nm      = 1.23984193*1.e-6/Etot_GeV # in nm
                #if parent_uid != 1: print(f"{time:.4f} photon    parent uid: {parent_uid}")
                self.data_buffer.photons = [parent_uid, pos_m, t_ns, dir, wl_nm]
            self.app.ph_counter += 1
            return G4ClassificationOfNewTrack.fKill
        elif Etot_GeV > self.app.E_skip_min and Etot_GeV <= self.app.E_skip_max and uid != 1 \
        and track.GetDefinition() in (G4Electron.Electron(), G4Positron.Positron()):
            parent_uid = track.GetParentID()
            pdgid = particle.GetPDGEncoding()
            pos_m = track.GetPosition()/m
            t_ns = track.GetGlobalTime()/ns
            dir = track.GetMomentumDirection()
            Etot_GeV = Etot_GeV
            step_length_m = track.GetStepLength()/m
            if self.app.save_process_name:
                process_name = track.GetCreatorProcess().GetProcessName()
                self.data_buffer.particles = [uid, parent_uid, pdgid, pos_m, t_ns,
                                              dir, Etot_GeV, process_name]
            else:
                self.data_buffer.particles = [uid, parent_uid, pdgid, pos_m, t_ns,
                                              dir, Etot_GeV]
            self.data_buffer.particles.counter += 1
            return G4ClassificationOfNewTrack.fKill
        elif Etot_GeV >= self.cerenkov_threshold_e and Etot_GeV >= self.app.E_lower_edge:
            parent_uid    = track.GetParentID()
            pdgid         = particle.GetPDGEncoding()
            pos_m         = track.GetPosition()/m
            t_ns          = track.GetGlobalTime()/ns
            Etot_GeV      = track.GetTotalEnergy()/GeV
            step_length_m = track.GetStepLength()/m
            if self.app.save_process_name:
                creator_process = track.GetCreatorProcess()
                if creator_process:
                    process_name = track.GetCreatorProcess().GetProcessName()
                else:
                    process_name = 'Init'
                self.data_buffer.tracks = [uid, parent_uid, pdgid, pos_m, t_ns,
                                           Etot_GeV, step_length_m, process_name]
            else:
                self.data_buffer.tracks = [uid, parent_uid, pdgid, pos_m, t_ns,
                                           Etot_GeV, step_length_m]
            return G4ClassificationOfNewTrack.fUrgent
        else:
            return G4ClassificationOfNewTrack.fKill


from g4camp.IO.gParticles import gParticles
from g4camp.IO.gPhotons import gPhotons
from g4camp.IO.gTracks import gTracks

class DataBuffer():

    def __init__(self):

        self._particles = gParticles('g4_particles')
        self._photons   = gPhotons('g4_photons')
        self._tracks    = gTracks('g4_tracks')

    @property
    def particles(self):
        return self._particles

    @particles.setter
    def particles(self, new_particle: list):
        self._particles.set_data(*new_particle)

    @property
    def photons(self):
        return self._photons

    @photons.setter
    def photons(self, new_photon: list):
        self._photons.set_data(*new_photon)

    @property
    def tracks(self):
        return self._tracks

    @tracks.setter
    def tracks(self, new_track: list):
        self._tracks.set_data(*new_track)

    def clear(self) -> None:
        self.__init__()


from g4camp.g4camp import g4camp
from time import time
import sys

n_events = 10
gun_args = { 'particle':   'e-', 
             'energy_GeV': 100,
             'position_m': [0,0,10],
             'direction':  [0,0,1]  }
optics = True

sim = g4camp(primary_generator='gun', gun_args=gun_args, optics=optics)
sim.setSkipMinMax('fraction', 0.0, 0.01)
sim.setPhotonSuppressionFactor(100)
sim.configure()

time0 = time()
for data in sim.run(n_events):
    particles = data.particles
    tracks = data.tracks
    photons = data.photons
    print(f"{photons.counter:>10}(x{sim.ph_suppression_factor}) photons, {tracks.counter:>6} tracks, {particles.counter:>6} particles;")
print(f"# Run time:  {(time()-time0):.2f} sec, {((time()-time0)/n_events):.2f} sec/event")
